/*
 *
 * Copyright (c) 2016-2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Adrian Szyndela <adrian.s@samsung.com>
 *          Łukasz Stelmach <l.stelmach@samsung.com>
 */
/*
 * @file crash-stack.h
 * @brief unwinding call stacks, header file for common functions
 */
#ifndef CRASH_STACK_H
#define CRASH_STACK_H

#define LOG_TAG "CRASH_STACK"
#include <stdint.h>
// #include <elfutils/libdwfl.h>
#include <json-c/json.h>
#include <libunwind.h>
#include <sys/procfs.h>

#define K_REP_VER         "report_version"
#define K_EXE_FILE_PATH   "exe_file_path"
#define K_SIGNAL          "signal"
#define K_MEM_INFO        "mem_info"
#define K_MI_MEMTOTAL     "mem_total"
#define K_MI_MEMFREE      "mem_free"
#define K_MI_BUFFS        "buffers"
#define K_MI_CACHED       "cached"
#define K_MI_VMPEAK       "vm_peak"
#define K_MI_VMSIZE       "vm_size"
#define K_MI_VMLCK        "vm_lck"
#define K_MI_VMPIN        "vm_pin"
#define K_MI_VMHWM        "vm_hwm"
#define K_MI_VMRSS        "vm_rss"
#define K_MI_VMDATA       "vm_data"
#define K_MI_VMSTK        "vm_stk"
#define K_MI_VMEXE        "vm_exe"
#define K_MI_VMLIB        "vm_lib"
#define K_MI_VMPTE        "vm_pte"
#define K_MI_VMSWAP       "vm_swap"
#define K_THREADS         "threads"
#define K_TH_PID          "pid"
#define K_TH_TID          "tid"
#define K_TH_TASKS        "tasks"
#define K_MAPS            "maps"
#define K_MAPS_START      "start"
#define K_MAPS_END        "end"
#define K_MAPS_PERM       "perm"
#define K_MAPS_PATH       "path"
#define K_REGISTERS       "registers"
#define K_REGS_NAME       "name"
#define K_REGS_VALUE      "value"
#define K_CALLSTACK       "callstack"
#define K_CALL_NAME       "name"
#define K_CALL_OFFSET     "offset"
#define K_CALL_ADDR       "addr"
#define K_CALL_M_NAME     "module_name"
#define K_CALL_M_OFFSET   "module_offset"
#define K_NATIVE_HASH     "native_hash"

#define REG_AARCH64_SP  31
#define REG_AARCH64_PC  32
#define REG_ARM_SP      13
#define REG_ARM_PC      12

#define ADD_REG_TO_JSON_ARRAY(j_array, name, value) \
	{ \
		json_object *j_reg = json_object_new_object(); \
		json_object_object_add(j_reg, "name", json_object_new_string(name)); \
		json_object_object_add(j_reg, "value", json_object_new_string(value)); \
		json_object_array_add(j_array, j_reg); \
	} while(0)

/**
 * @brief callstack.procedure info
 */
struct ProcInfo {
	uintptr_t addr; ///< procedure address
	int offset; ///< procedure offset
	char *name; ///< procedure name
	char *module_name; ///< module name
	int module_offset; ///< symbol's offset in the module
};
typedef struct ProcInfo ProcInfo;

#define MAX_CALLSTACK_LEN 1000
/**
 * @brief call stack database
 */
struct Callstack {
	ProcInfo proc[MAX_CALLSTACK_LEN]; ///< storage for elements
	size_t elems;						///< number of elements in the database
};
typedef struct Callstack Callstack;

/**
 * @brief Gets address of storage for specified register value
 *
 * @remarks This function uses static storage. Thus, it shouldn't be used in multiple threads.
 *
 * @param regname name of the register, e.g. "pc", "eax", "r7"
 * @param regnum ordinary number of the register, when name is not available
 * @return address of statically allocated storage of proper size for the register value
 */
void *_get_place_for_register_value(const char *regname, int regnum);

/**
 * @brief Creates call stack database.
 *
 * @param dwfl dwfl handle
 * @param pid pid of the analyzed process
 * @param callstack pointer to call stack database; results are put here
 * @param regs pointer to registers of the analyzed process
 */
void _create_crash_stack(void *dwfl, pid_t pid, Callstack *callstack, elf_gregset_t *regs);

/**
 * @brief Gets allocated space for getting target registers.
 *
 * @remarks This function uses static storage. Thus, it shouldn't be used in multiple threads.
 *
 * @param[out] size size of the registers structure
 * @return address of statically allocated storage of proper size for the registers
 */
void *_crash_stack_get_memory_for_registers(size_t *size);

/**
 * @brief Sets current values of registers from the buffer.
 *
 * @param regbuf buffer used for getting registers
 *
 * @see _crash_stack_get_memory_for_registers()
 */
void _crash_stack_set_registers(void *regbuf);

/**
 * @brief Return a value of a register from the buffer
 *
 * @param n libunwind register index, e.g. UNW_X86_EAX
 *
 * @return the value of the register
 */
unsigned long _get_register_value(int n);

/**
 * @brief Print values of registers to the JSON object
 *
 * @param jobj JSON object
 */
void _crash_stack_print_regs(json_object *jobj);

/**
 * @brief Return the set of ptrace(2)-free accessors for libunwind
 *
 * @return pointer to unw_accessors_t structure
 */
unw_accessors_t *_TB_accessors();

/**
 * @brief Create a context for analysis
 *
 * @param pid pid of the analyzed process
 *
 * @return a pointer to pass as the third argument for unw_inint_remote()
 */
void *_TB_create(pid_t pid);

/**
 * @brief Free any resources allocated by _TB_create()
 *
 * @param ptr pointer returned by _TB_create()
 */
void _TB_destroy(void *ptr);

/**
 * @brief Find the name of the module and the offset for the address
 *
 * @param addr the address of the code to find the module for
 * @param bufp the output buffer for the module name
 * @param buf_len the size of the buffer
 * @param offp output the offset of the code inside the module
 *
 * @return zero on success
 */
int get_mod_name(unw_word_t addr, char *bufp,
		 size_t buf_len, unw_word_t *offp);
#endif /* CRASH_STACK_H */
