/* -*- mode: C; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Copyright (c) 2014, Tbricks AB
 * Copyright (c) 2018 Samsung Electronics Co., Ltd.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL TBRICKS
 * AB BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THISS OFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <libgen.h>
#include <libunwind.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/user.h>
#include <unistd.h>

#include "dwarf.h"
#include "mem_map.h"
#include "proc.h"
#include "crash-stack.h"

#define LOG_TAG "CRASH_STACK"
#include "shared/log.h"
#include "shared/telf.h"

size_t stack_size = 0xa00000;

/*
 * some libelf implementations do not provide ELF_C_READ_MMAP
 */
#if HAVE_DECL_ELF_C_READ_MMAP
#define TBSTACK_ELF_C_READ ELF_C_READ_MMAP
#else
#define TBSTACK_ELF_C_READ ELF_C_READ
#endif

#define to_unw_word(p) ((unw_word_t) (uintptr_t) (p))

/*
 * search unwind table for a procedure (used by find_proc_info)
 */
#if defined(__arm__)
#define search_unwind_table UNW_OBJ(search_unwind_table)
#else
#define search_unwind_table UNW_OBJ(dwarf_search_unwind_table)
#endif
extern int search_unwind_table(unw_addr_space_t as, unw_word_t ip,
        unw_dyn_info_t *di, unw_proc_info_t *pip,
        int need_unwind_info, void *arg);

static struct mem_map *map = NULL;
static unsigned long eip = 0;
static unsigned long esp = 0;
static FILE *debug;


static Elf *elf_start(int fd, char *image, uint64_t size)
{
    Elf *elf = malloc(sizeof(Elf));

    if (fd > 0) {
        if (!teu_begin(fd, elf))
            _E("elf_begin: %s", teu_errmsg(elf));
    } else {
        if (!teu_begin_memory(image, elf))
            _E("elf_memory: %s", teu_errmsg(elf));
    }

    return elf;
}

#if defined(__arm__)
/*
 * find section .ARM.exidx in ELF binary
 */
static int find_exidx(int fd, char *image, uint64_t size,
		      uint64_t *table_data, uint64_t *table_len)
{
    Elf *elf;
    uint64_t offset = 0;

    if ((elf = elf_start(fd, image, size)) == NULL)
        return -1;

    for (size_t i = 0; i < elf->ehdr.e_shnum; i++) {
        Elf64_Shdr shdr;

        if (!teu_getshdr(elf, i, &shdr)) {
            _E("Can not read section header: %s", teu_errmsg(elf));
            break;
        }

        if (shdr.sh_type == SHT_ARM_EXIDX) {
            offset = *table_data = shdr.sh_offset;
	    *table_len = shdr.sh_size;
        }
    }

    teu_close(elf);
    free(elf);
    return (offset ? 0 : -1);
}
#else /* __arm__ */
/*
 * get dwarf encoded value
 */
static ssize_t dw_get_value(char *data, unsigned char enc,
        uint64_t cur, uint64_t *value)
{
    int64_t number = 0;
    size_t size;

    if (enc == DW_EH_PE_omit) {
        *value = 0;
        return 0;
    }

    if (enc == DW_EH_PE_absptr) {
        *value = *(uint64_t *)data;
        return 8;
    }

    switch (enc & 0xf)
    {
    case DW_EH_PE_udata2:
        number = *(uint16_t *)data;
        size = 2;
        break;

    case DW_EH_PE_sdata2:
        number = *(int16_t *)data;
        size = 2;
        break;

    case DW_EH_PE_udata4:
        number = *(uint32_t *)data;
        size = 4;
        break;

    case DW_EH_PE_sdata4:
        number = *(int32_t *)data;
        size = 4;
        break;

    case DW_EH_PE_udata8:
        number = *(uint64_t *)data;
        size = 8;
        break;

    case DW_EH_PE_sdata8:
        number = *(int64_t *)data;
        size = 8;
        break;

    default:
        _E("unsupported encoding in "
                ".eh_frame_hdr: %d", (int)enc);
        return -1;
    }

    switch (enc & 0xf0) {
    case DW_EH_PE_absptr:
        *value = number;
        break;

    case DW_EH_PE_pcrel:
        *value = cur + number;
        break;

    default:
        return -1;
    }

    return size;
}

/*
 * parse contents of .eh_frame_hdr
 */
static int parse_eh_frame_hdr(char *data, size_t pos,
        uint64_t *table_data, uint64_t *fde_count)
{
    char version, eh_frame_ptr_enc, fde_count_enc;
    ssize_t size;
    uint64_t eh_frame_ptr;

    version = data[0];
    eh_frame_ptr_enc = data[1];
    fde_count_enc = data[2];
    data += 4;
    pos += 4;

    if (version != 1) {
        _E("unknown .ehf_frame_hdr version %d", version);
        return -1;
    }

    size = dw_get_value(data, eh_frame_ptr_enc, pos, &eh_frame_ptr);
    if (size < 0)
        return -1;
    pos += size;
    data += size;

    size = dw_get_value(data, fde_count_enc, pos, fde_count);
    if (size < 0)
        return -1;
    pos += size;
    *table_data = pos;

    return 0;
}

/*
 * find section .eh_frame_hdr in ELF binary
 */
static int find_eh_frame_hdr(int fd, char *image, uint64_t size,
        uint64_t *table_data, uint64_t *segbase, uint64_t *fde_count)
{
    Elf *elf;
    Elf64_Ehdr ehdr;
    Elf64_Shdr shdr;
    uint64_t offset = 0;

    if ((elf = elf_start(fd, image, size)) == NULL)
        return -1;

    if (!teu_getehdr(elf, &ehdr)) {
        _E("teu_getehdr: %s", teu_errmsg(elf));
        goto elf_section_offset_end;
    }

    for (size_t i = 0; i < elf->ehdr.e_shnum; i++) {
        if (!teu_getshdr(elf, i, &shdr))
            break;

        char *str = teu_strptr(elf, ehdr.e_shstrndx, shdr.sh_name);
        if (str != NULL && !strcmp(str, ".eh_frame_hdr")) {
            void *data = teu_getsdata(elf, &shdr);

            offset = *segbase = shdr.sh_offset;
            parse_eh_frame_hdr(data, offset, table_data, fde_count);
            break;
        }
    }

elf_section_offset_end:
    teu_close(elf);
    free(elf);
    return (offset ? 0 : -1);
}
#endif /* __arm__ */

static int find_unwind_table(int fd, char *image, uint64_t size,
			     uint64_t *table_data, int *table_format,
			     uint64_t *segbase, uint64_t *table_len)
{
#if defined(__arm__)
	(void) segbase;
	*table_format = UNW_INFO_FORMAT_ARM_EXIDX;
	return find_exidx(fd, image, size, table_data, table_len);
#else
	(void) table_data;
	*table_format = UNW_INFO_FORMAT_REMOTE_TABLE;
	return find_eh_frame_hdr(fd, image, size, table_data, segbase, table_len);
#endif
}


/*
 * dynamic array of symbols
 */
struct symbols
{
    Elf64_Sym *s_data;
    size_t s_size;
    size_t s_cap;
};

/*
 * add a symbol to array
 */
static int push_symbol(struct symbols *array, const Elf64_Sym *s)
{
    ++array->s_size;
    if (array->s_size > array->s_cap) {
        Elf64_Sym *new_data;
        array->s_cap <<= 1;
        new_data = malloc(sizeof(Elf64_Sym) * array->s_cap);
        if (new_data == NULL) {
            _E("malloc(): %m");
            return -1;
        }
        memcpy(new_data, array->s_data, sizeof(Elf64_Sym) * (array->s_size-1));
        free(array->s_data);
        array->s_data = new_data;
    }
    memcpy(array->s_data + (array->s_size-1), s, sizeof(Elf64_Sym));
    return 0;
}

/*
 * symbol comparison function for qsort
 */
static int sym_compar(const void *v1, const void *v2)
{
    const Elf64_Sym *s1 = v1;
    const Elf64_Sym *s2 = v2;

    if (s1->st_value < s2->st_value)
        return -1;
    if (s1->st_value > s2->st_value)
        return 1;
    return 0;
}

/*
 * get function name
 *
 * fd: open binary
 * load: mmap address
 * offset: file offset
 * addr: ip value
 * off: offset within the function
 */
static char *proc_name(int fd, char *image, size_t size, uint64_t load,
        uint64_t offset, uint64_t addr, unw_word_t *off)
{
    Elf *elf;
    char *str = NULL;
    int rc = 0;
    struct symbols all;
    size_t pnum, i;
    uint64_t vaddr = 0;

    /*
     * open ELF handle
     */
    if ((elf = elf_start(fd, image, size)) == NULL)
        return NULL;

    /*
     * initialize dynamic array
     */
    all.s_cap = 64;
    all.s_size = 0;
    all.s_data = malloc(all.s_cap * sizeof(Elf64_Sym));

    if (all.s_data == NULL)
	    goto proc_name_end;

    pnum = elf->ehdr.e_phnum;
    if (pnum == 0)
        goto proc_name_end;

    for (i = 0; i < pnum; ++i) {
        Elf64_Phdr phdr;
        if (!teu_getphdr(elf, i, &phdr))
            goto proc_name_end;
        if (phdr.p_type != PT_LOAD)
            continue;
        if (phdr.p_flags != (PF_X | PF_R))
            continue;
        if ((phdr.p_offset & ~(phdr.p_align - 1)) != offset)
            continue;
        vaddr = phdr.p_vaddr;
        break;
    }

    /*
     * adjust address
     */
    addr -= load;
    addr += offset;

    /*
     * search symtab or dynsym section
     */
    for (size_t i = 0; i < elf->ehdr.e_shnum; i++) {
        Elf64_Shdr shdr;

        if (!teu_getshdr(elf, i, &shdr)) {
            goto proc_name_end;
        }

        if (shdr.sh_type == SHT_DYNSYM || shdr.sh_type == SHT_SYMTAB) {
            int symbol_count;

            void *data = teu_getsdata(elf, &shdr);

            if (shdr.sh_entsize == 0)
                symbol_count = 0;
            else
                symbol_count = shdr.sh_size / shdr.sh_entsize;

            for (i = 0; i < (size_t)symbol_count; ++i) {
                Elf64_Sym s;

                if (!teu_getsym(elf, &shdr, i, data, &s)) {
                    _E("teu_getsym: %s", teu_errmsg(elf));
                    rc = -1;
                    goto proc_name_end;
                }

                if (ELF64_ST_TYPE(s.st_info) != STT_FUNC)
                    continue;

                /*
                 * adjust symbol value
                 */
                s.st_value -= vaddr;

                /*
                 * exact match
                 */
                if (addr >= s.st_value && addr < (s.st_value + s.st_size)) {
                    str = teu_strptr(elf, shdr.sh_link, s.st_name);
                    if (str == NULL) {
                        _E("teu_strptr #1: %s", teu_errmsg(elf));
                        rc = -1;
                        goto proc_name_end;
                    }
                    str = strdup(str);
                    *off = addr - s.st_value;
                    goto proc_name_end;
                }

                /* store section link */
                s.st_shndx = shdr.sh_link;
                /*
                 * save symbol in array
                 */
                if (push_symbol(&all, &s) < 0)
                    goto proc_name_end;
            }
        }
    }

    /*
     * sometimes function symbols have zero size but contain the code.
     * common example is _start on most systems.
     * in this case we try to find two adjacent symbols with first
     * one of zero size
     */
    if (!rc && str == NULL) {
        qsort(all.s_data, all.s_size, sizeof(Elf64_Sym), sym_compar);
        for (i = 0; i < (all.s_size-1); ++i) {
            const Elf64_Sym *cur = all.s_data + i;
            const Elf64_Sym *next = all.s_data + i + 1;
            if (cur->st_size == 0) {
                if (cur->st_value <= addr && addr < next->st_value) {
                    str = teu_strptr(elf, cur->st_shndx, cur->st_name);
                    if (str == NULL) {
                        _E("teu_strptr #2: %s", teu_errmsg(elf));
                        rc = -1;
                        goto proc_name_end;
                    }
                    str = strdup(str);
                    *off = addr - cur->st_value;
                    goto proc_name_end;
                }
            }
        }
    }

proc_name_end:
    free(all.s_data);
    teu_close(elf);
    free(elf);
    return str;
}

/*
 * get mmapped ELF image info
 */
static int get_elf_image_info(struct mem_region *region,
        char **elf_image, uint64_t *elf_length, uintptr_t ip)
{
    struct mem_data_chunk *chunk;

    if ((chunk = mem_region_find_data_chunk(region, (void *)ip)) == NULL)
        return -1;

    if (chunk->data == NULL)
        return -1;

    *elf_image = chunk->data;
    *elf_length = chunk->length;

    return 0;
}

/*
 * find unwind info for function
 */
static int find_proc_info(unw_addr_space_t as, unw_word_t ip,
        unw_proc_info_t *pip, int need_unwind_info, void *arg)
{
    struct mem_region *region;
    char *elf_image = NULL;
    uint64_t elf_length = 0;
    unw_dyn_info_t di;
    uint64_t table_data = 0;
    uint64_t segbase, fde_count;
    int table_format = -1;
    int rc = -UNW_EINVAL;

    (void) arg;

    if (ip == 0)
        return -UNW_ENOINFO;

    if ((region = mem_map_get_file_region(map, (void *)ip)) == NULL)
        return rc;

    if (region->fd < 0 && region->type != MEM_REGION_TYPE_VDSO
        && region->type != MEM_REGION_TYPE_VSYSCALL)
        return rc;

    if (region->fd < 0 &&
        get_elf_image_info(region, &elf_image, &elf_length, ip) < 0)
        return rc;

    memset(&di, 0, sizeof(di));

    if (find_unwind_table(region->fd, elf_image, elf_length,
                          &table_data, &table_format, &segbase,
                          &fde_count) !=0) {
        rc = -UNW_ENOINFO;
    } else {
        di.format = table_format;
        di.start_ip = (unw_word_t)region->start;
        di.end_ip = (unw_word_t)region->start + region->length;

        if (table_format == UNW_INFO_FORMAT_REMOTE_TABLE) {
            di.u.rti.name_ptr = 0;
            di.u.rti.segbase = (unw_word_t)(region->start - region->offset) + segbase;
            di.u.rti.table_data = (unw_word_t)(region->start - region->offset) + table_data;
            di.u.rti.table_len = fde_count * sizeof(uint32_t) * 2 / sizeof(unw_word_t);
        } else if (table_format == UNW_INFO_FORMAT_ARM_EXIDX) {
            di.u.rti.name_ptr = to_unw_word(region->path);
            di.u.rti.table_data = (unw_word_t)region->start + table_data;
            di.u.rti.table_len = fde_count;
        }
        rc = search_unwind_table(as, ip, &di, pip, need_unwind_info, arg);
    }

    return rc;
}

/*
 * put_unwind_info: do nothing
 */
static void put_unwind_info(unw_addr_space_t as,
        unw_proc_info_t *pip, void *arg)
{
    (void) as;
    (void) pip;
    (void) arg;
}

/*
 * not used
 */
static int get_dyn_info_list_addr(unw_addr_space_t as,
        unw_word_t *dilap, void *arg)
{
    (void) as;
    (void) dilap;
    (void) arg;
    return -UNW_ENOINFO;
}

/*
 * read a word from memory. we use mem_map for that
 */
static int access_mem(unw_addr_space_t as, unw_word_t addr,
        unw_word_t *valp, int write, void *arg)
{
    (void) as;
    (void) arg;

    if (write) {
        _E("access_mem: requested write, rejecting");
        return -UNW_EINVAL;
    }

    return mem_map_read_word(map, (void *)(uintptr_t)addr, valp);
}

/*
 * get register value
 */
static int access_reg(unw_addr_space_t as, unw_regnum_t reg,
        unw_word_t *val, int write, void *arg)
{

	(void) as;
	(void) val;
	(void) write;
	(void) arg;


	if (write) {
		_E("requested to write into register");
		return -UNW_EINVAL;
	}

	*val = _get_register_value(reg);
	return 0;
}

/*
 * __cxa_demangle() is taken from libstdc++, however there is no header that we
 * can take a declaration from. Importing through 'extern' allows using it.
 */
/// @cond false
extern char *__cxa_demangle(const char *mangled_name, char *output_buffer,
		size_t *length, int *status);
///@endcond

/**
 * @brief Checks if symbol starts with '_Z' prefix
 *
 * @param symbol string to compare
 */
static int is_symbol_demanglable(const char *symbol)
{
	return symbol != 0 && (strlen(symbol) >= 2) &&
		symbol[0] == '_' && symbol[1] == 'Z';
}

/**
 * @brief Replaces symbols with demangled
 *
 * @param proc_info gathered call stack element
 */
static void demangle_symbol(char **symbol)
{
	int status = -1;
	char *demangled_symbol = NULL;
	if (!is_symbol_demanglable(*symbol))
		return;
	demangled_symbol = __cxa_demangle(*symbol, NULL, NULL, &status);
	if (status == 0) {
		free(*symbol);
		*symbol = demangled_symbol;
	}
	return;
}

/**
 * @brief Find the name of the module and the offset for the address
 *
 * @param addr the address of the code to find the module for
 * @param bufp the output buffer for the module name
 * @param buf_len the size of the buffer
 * @param offp output the offset of the code inside the module
 *
 * @return zero on success
 */
int get_mod_name(unw_word_t addr, char *bufp,
		 size_t buf_len, unw_word_t *offp)
{
	struct mem_region *region;

	region = mem_map_get_file_region(map, (void *)addr);
	if (region == NULL || region->path == NULL || region->path[0] == '\0')
		return -1;

	strncpy(bufp, region->path, buf_len);
	bufp[buf_len - 1] = '\0';
	*offp = addr - (uintptr_t)region->start;
	return 0;
}

/*
 * floating point registers are not used
 */
static int access_fpreg(unw_addr_space_t as, unw_regnum_t regnum,
        unw_fpreg_t *fpvalp, int write, void *arg)
{
    (void) as;
    (void) regnum;
    (void) fpvalp;
    (void) write;
    (void) arg;

    _D("access_fpreg is not supported");
    return -UNW_ENOINFO;
}

/*
 * not used
 */
static int resume(unw_addr_space_t as, unw_cursor_t *cp, void *arg)
{
    (void) as;
    (void) cp;
    (void) arg;

    _D("resume is not supported");
    return -UNW_ENOINFO;
}

/*
 * get function name callback
 */
static int get_proc_name(unw_addr_space_t as, unw_word_t addr, char *bufp,
        size_t buf_len, unw_word_t *offp, void *arg)
{
    struct mem_region *region;
    char *name = NULL;

    (void) as;
    (void) arg;

    if (addr == 0)
        return -UNW_ENOINFO;

    if ((region = mem_map_get_file_region(map, (void *)addr)) == NULL)
        return -UNW_ENOINFO;

    if (region->fd < 0 && region->type == MEM_REGION_TYPE_DELETED) {
        const char *base = basename(region->path);
        snprintf(bufp, buf_len, "?? (%s is deleted)", base);
        *offp = 0;
        return 0;
    } else if (region->type == MEM_REGION_TYPE_MMAP ||
            region->type == MEM_REGION_TYPE_VDSO ||
            region->type == MEM_REGION_TYPE_VSYSCALL) {
        char *elf_image = NULL;
        uint64_t elf_length = 0;

        if (region->fd < 0 &&
                get_elf_image_info(region, &elf_image, &elf_length, addr) < 0)
            return -UNW_ENOINFO;

        name = proc_name(region->fd, elf_image, elf_length,
                (uint64_t)(uintptr_t)region->start, region->offset, addr, offp);
        demangle_symbol(&name);
    }

    if (name == NULL) {
        /*
         * if name cannot be resolved, print binary file name or region type
         */
        if (region->type == MEM_REGION_TYPE_MMAP) {
	    return -UNW_ENOINFO;
        } else {
            snprintf(bufp, buf_len, "?? [%s]",
                 str_mem_region_type(region->type));
        }
        *offp = 0;
        return 0;
    }

    strncpy(bufp, name, buf_len);
    free(name);

    return 0;
}

/*
 * libunwind remote callbacks
 */
static unw_accessors_t snapshot_addr_space_accessors = {
    .find_proc_info = find_proc_info,
    .put_unwind_info = put_unwind_info,
    .get_dyn_info_list_addr = get_dyn_info_list_addr,
    .access_mem = access_mem,
    .access_reg = access_reg,
    .access_fpreg = access_fpreg,
    .resume = resume,
    .get_proc_name = get_proc_name,
};

unw_accessors_t *_TB_accessors() {
	return &snapshot_addr_space_accessors;
}

void _get_esp_and_eip() {
    int esp_i, eip_i;
#if defined(__x86_64__)
    esp_i = UNW_X86_64_RSP;
    eip_i = UNW_X86_64_RIP;
#elif defined(__i386__)
    esp_i = UNW_X86_ESP;
    eip_i = UNW_X86_EIP;
#elif defined(__arm__)
    esp_i = REG_ARM_SP; // values from crash-stack-x86.c
    eip_i = REG_ARM_PC;
#elif defined(__aarch64__)
    esp_i = REG_AARCH64_SP; // values from crash-stack-aarch64.c
    eip_i = REG_AARCH64_PC;
#else
#error "Not supported architecture"
#endif
    esp = _get_register_value(esp_i);
    eip = _get_register_value(eip_i);
}


void *_TB_create (pid_t pid) {
#define STAT_LINE_MAXSIZE 4096
	int n_frames;
	struct mem_data_chunk **stacks_cover = NULL;
	int rc;
	void *ret = NULL;
	char *buf = NULL;
	char *p = NULL;
	FILE *f = NULL;
	int i;

	long page;
	long label;

	if ((page = sysconf(_SC_PAGESIZE)) < 0) {
		_E("sysconf(_SC_PAGESIZE): %m");
		return NULL;
	}
	--page;

	map = create_maps(pid);

	buf = malloc(STAT_LINE_MAXSIZE);
	if (!buf)
		return NULL;

	debug = stderr;

	snprintf(buf, STAT_LINE_MAXSIZE, "/proc/%d/stat", pid);
	f = fopen(buf, "r");
	if (!f)
		goto out_err;

	if (fgets(buf, STAT_LINE_MAXSIZE, f) == NULL)
		goto out_err;

	// The filename of the executable (the second field) is in parentheses and can
	// contain whitespaces. Therefore I'm looking for the last ')' character
	// and then iterate to the 27 field from here.
	p = strrchr(buf, ')');
	if (!p)
		goto out_err;
	for (i = 0; i < 27; i++) {
		p = strchr(p, ' ');
		if (!p)
			goto out_err;
		p++;
	}

	_get_esp_and_eip();

	ret = (void*)&snapshot_addr_space_accessors;

	label = esp & ~page;
	mem_map_add_label(map, (void *)label, 1);

	/*
	 * arrange data chunks to copy memory contents. in most cases the chunks
	 * will start from %rsp pointing somewhere in thread's stack
	 * to the end of the stack region
	 */
	stacks_cover = malloc(sizeof(struct mem_data_chunk*) * 1);
	if (stacks_cover == NULL) {
		fprintf(stderr, "error: stacks cover == NULL\n");
		goto out_err;
	}

	n_frames = mem_map_build_label_cover(map, stack_size,
					     stacks_cover, page + 1);

	/*
	 * copy memory contents
	 */
	rc = copy_memory(pid, stacks_cover, n_frames);

	if (rc < 0)
		ret = NULL;

out_err:
        if (stacks_cover)
		free(stacks_cover);
	if (f)
		fclose(f);
	if (buf)
		free(buf);
	return ret;
}

void _TB_destroy (void *ptr) {
	(void) ptr;
        if (debug)
		fprintf(debug, "_TB_destroy\n");
	if (debug && debug != stdout && debug != stderr)
		fclose(debug);
}
