/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#ifndef CORE_HPP__
#define CORE_HPP__

#include "helpers.hpp"
#include "log.hpp"
#include "note.hpp"
#include "program.hpp"
#include "shared/telf.h"

#include <fcntl.h>
#include <inttypes.h>
#include <limits.h>
#include <link.h>
#include <sys/procfs.h>

extern "C" {
#include <thread_db.h>
}

#include <cstring>
#include <iostream>
#include <map>
#include <memory>
#include <string>
#include <vector>

static constexpr size_t PAGESIZE = 0x1000;

template <typename T>
class Core {
 private:
	ProgramTableEntryBase<T> *m_note_header;
	std::vector<std::unique_ptr<ProgramTableEntryBase<T>>> m_phdrt;
	std::vector<std::unique_ptr<SymData>> m_symdata;
	size_t current_data_address;
	std::vector <std::unique_ptr<Note>> m_notes;
	typename T::Ehdr m_ehdr;

 public:
	Core() : m_note_header(new ProgramTableEntryNote<T>()),
	         current_data_address(sizeof(m_ehdr)),
	         m_ehdr() {
	         m_phdrt.push_back(std::unique_ptr<ProgramTableEntryBase<T>>(m_note_header));
	}

	void FillELFHeader() {
		memcpy(m_ehdr.e_ident, ELFMAG, sizeof(ELFMAG));
		m_ehdr.e_ident[EI_CLASS] = T::ELFCLASS;
		m_ehdr.e_ident[EI_DATA] = ELFDATA2LSB;
		m_ehdr.e_ident[EI_VERSION] = EV_CURRENT;
		m_ehdr.e_ident[EI_OSABI] = ELFOSABI_SYSV;
		m_ehdr.e_ident[EI_ABIVERSION] = 0;
		memset(&m_ehdr.e_ident[EI_PAD], 0, sizeof(m_ehdr.e_ident) - EI_PAD);
		m_ehdr.e_type = ET_CORE;
		m_ehdr.e_machine = MACHINE;
		m_ehdr.e_version = EV_CURRENT;
		m_ehdr.e_entry = 0;
		m_ehdr.e_phoff = sizeof(m_ehdr);
		m_ehdr.e_shoff = 0;
		m_ehdr.e_flags = 0;
		m_ehdr.e_ehsize = sizeof(m_ehdr);
		m_ehdr.e_phentsize = sizeof(typename T::Phdr);
		m_ehdr.e_phnum = m_phdrt.size();
		m_ehdr.e_shentsize = 0;
		m_ehdr.e_shnum = 0;
		m_ehdr.e_shstrndx = 0;
	}

	void ImportMapRecord(std::shared_ptr<MapRecord> record) {
		ProgramTableEntryLoad<T> *entry = new ProgramTableEntryLoad<T>();
		entry->header.p_flags = static_cast<decltype(entry->header.p_flags)>(record->perms);
		entry->header.p_offset = current_data_address;
		current_data_address += record->end - record->start;
		entry->header.p_vaddr = record->start;
		entry->header.p_paddr = 0;
		entry->header.p_filesz = record->end - record->start;
		entry->header.p_memsz = record->end - record->start;
		entry->header.p_align = PAGESIZE;
		entry->important = record->important;
		m_phdrt.push_back(std::unique_ptr<ProgramTableEntryLoad<T>>(entry));
	}

	void SaveNotes(std::ofstream &core_file) {
		for (auto const &note : m_notes)
			note->SaveTo(core_file);
	}

	void SaveELFHeader(std::ofstream &core_file) {
		FillELFHeader();
		core_file.write(reinterpret_cast<char*>(&m_ehdr), sizeof(m_ehdr));
	}

	void SaveProgramHeadersTable(std::ofstream &core_file) {
		for (const auto& record : m_phdrt) {
			core_file.write(reinterpret_cast<char*>(&record->header),
							  sizeof(record->header));
		}
	}

	bool IsEmpty(char *buff, size_t size) {
		uint32_t *tmp_buff = reinterpret_cast<uint32_t*>(buff);
		const size_t divider = sizeof(tmp_buff[0])/sizeof(buff[0]);

		assert(size % divider == 0);

		for (size_t i = 0; i < size/divider; i++) {
			if (tmp_buff[i] != 0)
				return false;
		}
		return true;
	}

	void CopyData(int input, std::ofstream &output, size_t size) {
		off64_t len = PAGESIZE;
		char buff[PAGESIZE];

		while (size > 0) {
			if (size < len)
				len = size;

			off64_t cur_pos = lseek64(input, 0, SEEK_CUR);
			if (cur_pos == -1) {
				logger.log_error("lseek error: %s", std::system_category().default_error_condition(errno).message().c_str());
				return;
			}

			ssize_t bytes_read;
			if ((bytes_read = read(input, buff, len)) == -1) {
				// tell() returns 128bits value so I cast this to 64bit to be able to print it
				logger.log_error("read error at 0x%llx: %s\n", static_cast<long long int>(output.tellp()),
						std::system_category().default_error_condition(errno).message().c_str());
				lseek64(input, cur_pos + len, SEEK_SET);
				output.seekp(len, std::ios_base::cur);
				size -= len;
			} else {
				if (bytes_read == PAGESIZE && IsEmpty(buff, len)) {
					output.seekp(bytes_read, std::ios_base::cur);
				} else {
					output.write(buff, bytes_read);
				}
				size -= bytes_read;
			}
		}
	}

	void SaveLoadable(int mem_fd, std::ofstream &core_file, bool minicore) {
		for (const auto &record : m_phdrt) {
			if (record->header.p_type != PT_LOAD)
				continue;
			if (lseek64(mem_fd, record->header.p_vaddr, SEEK_SET) == -1)
				logger.log_error("lseek64 error: %s", std::system_category().default_error_condition(errno).message().c_str());
			if (!minicore || static_cast<ProgramTableEntryLoad<T>*>(record.get())->important)
				CopyData(mem_fd, core_file, record->header.p_filesz);
			else
				core_file.seekp(record->header.p_filesz, std::ios_base::cur);
		}
	}

	void AddNote(Note *note) {
		m_note_header->header.p_filesz += note->GetSize();
		m_note_header->header.p_memsz += note->GetSize();
		m_notes.push_back(std::unique_ptr<Note>(note));
	}

	void AddPRSTATUSNote(const user_regs_struct &regs, const pid_t pid) {
		NoteCorePRStatus *note = new NoteCorePRStatus(regs, pid);
		AddNote(note);
	}

	void AddFILENote(const std::vector<std::shared_ptr<MapRecord>> &records) {
		NoteCoreFile<T> *note = new NoteCoreFile<T>();
		note->ImportRecords(records);
		AddNote(note);
	}

	void AddAUXVNote(const std::vector<char> &auxv_vector) {
		NoteCoreAUXV *note = new NoteCoreAUXV();
		note->SetAuxv(auxv_vector);
		AddNote(note);
	}

	void ImportMapRecords(const std::vector<std::shared_ptr<MapRecord>> &records) {
		current_data_address += (records.size() + 1)*sizeof(typename T::ProgramTableHeader);
		m_note_header->header.p_offset = current_data_address;

		for (const auto &note : m_notes)
			current_data_address += note->GetSize();

		for (const auto &record : records)
			ImportMapRecord(record);
	}

	typename T::uint AUXVval(const typename T::auxv_t *auxv, typename T::Addr type) {
		for (int i = 0; auxv[i].a_type != AT_NULL; i++) {
			if (auxv[i].a_type == type)
				return auxv[i].a_un.a_val;
		}
		return 0;
	}

	void ReadFromFile(std::ifstream &input, unsigned long address, void *data, size_t len) {
		input.seekg(address, std::ios_base::beg);
		input.read(static_cast<char*>(data), len);
	}

	void ReadFromFile(const int fd, unsigned long address, void *data, size_t len) {
		if (lseek64(fd, address, SEEK_SET) == -1)
			throw std::system_error(errno, std::system_category(), "failed to lseek64 at " + std::to_string(address));
		if (read(fd, data, len) == -1)
			throw std::system_error(errno, std::system_category(), "failed to read at " + std::to_string(address));
	}

	void DumpData(int fd, std::ofstream &output, typename T::Addr iaddress, typename T::Addr oaddress, size_t len, const std::string &desc) {
		if (lseek64(fd, iaddress, SEEK_SET) == -1) {
			logger.log_info("failed to lseek64");
			return;
		}

		output.seekp(oaddress, std::ios_base::beg);
		constexpr const char * format = std::is_same<T, Elf64>::value ?
		                     "dumping %s: 0x%" PRIx64 "-0x%" PRIx64 " to 0x%" PRIx64 " (%zu bytes)" :
		                     "dumping %s: 0x%" PRIx32 "-0x%" PRIx32 " to 0x%" PRIx32 " (%zu bytes)";

		logger.log_info(format, desc.c_str(), iaddress, iaddress+len, oaddress, len);
		CopyData(fd, output, len);
	}

	void DumpData(int fd, std::ofstream &output, typename T::Addr iaddress, size_t len, const std::string &desc) {
		typename T::Addr out_addr = MemAddr2FileAddr(iaddress);
		if (out_addr)
			DumpData(fd, output, iaddress, out_addr, len, desc);
		else
			logger.log_error("No address for 0x%x", iaddress);
	}

	ProgramTableEntryLoad<T>* FindRecordForMemAddress(typename T::Addr address) {
		for (auto &record : m_phdrt) {
			if (record->header.p_type != PT_LOAD)
				continue;

			ProgramTableEntryLoad<T>* entry = static_cast<ProgramTableEntryLoad<T>*>(record.get());
			if (address >= entry->header.p_vaddr && address < entry->header.p_vaddr + entry->header.p_memsz) {
				return entry;
			}
		}
		return nullptr;
	}

	typename T::Addr MemAddr2FileAddr(typename T::Addr address) {
		ProgramTableEntryLoad<T> *record = FindRecordForMemAddress(address);

		if (record == nullptr)
			return 0;

		return record->header.p_offset + (address - record->header.p_vaddr);
	}

	SymData* AllocSymData(const std::string &lib, unsigned long start, typename T::Word type) {
		auto sym_data = new SymData();

		sym_data->start = start;
		sym_data->fd = open(lib.c_str(), O_RDONLY);

		if (sym_data->fd < 0) {
			delete sym_data;
			return nullptr;
		}

		sym_data->elf = new Elf();
		teu_begin(sym_data->fd, sym_data->elf);

		for (size_t i = 0; i < sym_data->elf->ehdr.e_shnum; i++) {
			if (teu_getshdr(sym_data->elf, i, &sym_data->shdr) &&
			                 sym_data->shdr.sh_type == type)
				break;
		}

		sym_data->data = teu_getsdata(sym_data->elf, &sym_data->shdr);
		if (sym_data->shdr.sh_entsize == 0) {
			logger.log_info("sh_entsize is zero for %s", lib.c_str());
			teu_close(sym_data->elf);
			delete sym_data->elf;
			close(sym_data->fd);
			delete sym_data;
			return nullptr;
		}
		sym_data->count = sym_data->shdr.sh_size / sym_data->shdr.sh_entsize;

		return sym_data;
	}

	void SaveSymData(const std::string &lib, unsigned long start) {
		SymData* sym_data;

		for (auto type: {SHT_SYMTAB, SHT_DYNSYM}) {
			sym_data = AllocSymData(lib, start, type);
			if (!sym_data)
				continue;

			m_symdata.push_back(std::unique_ptr<SymData>(sym_data));
		}
	}

	static int FindPthreadsCb(const td_thrhandle_t *th, void *cb_data) {
		td_thrinfo_t thinfo;
		td_thr_get_info(th, &thinfo);
		return TD_OK;
	}

	void SavePthreadList(int mem_fd, std::ofstream &core_file, int pid) {
		struct ps_prochandle ph = {
			.core_obj = this,
			.bit32 = (std::is_same<T, Elf32>::value),
			.mem_fd = mem_fd,
			.core_file = core_file,
			.pid = pid
		};

		td_thragent_t *ta;
		td_err_e err;

		err = td_ta_new(&ph, &ta);
		if (err == TD_OK) {
			err = td_ta_thr_iter(ta, FindPthreadsCb, nullptr, TD_THR_ANY_STATE, TD_THR_LOWEST_PRIORITY,
			                     TD_SIGNO_MASK, TD_THR_ANY_USER_FLAGS);
			td_ta_delete(ta);
		}

		if (err == TD_NOLIBTHREAD) {
			logger.log_info("target does not appear to be multi-threaded");
		} else if (err != TD_OK) {
			logger.log_info("WARNING: libthread_db not found, using fallback");
			// TODO: fallback
		}
	}

	bool SymAddress(const char *sym_name, unsigned long *addr) {
		for (const auto &sd : m_symdata) {
			for (int i = 0; i < sd->count; i++) {
				Elf64_Sym sym;

				if (!teu_getsym(sd->elf, &sd->shdr, i, sd->data, &sym))
					continue;

				const char *st = teu_strptr(sd->elf, sd->shdr.sh_link, sym.st_name);

				if (strcmp(st, sym_name) != 0)
					continue;

				*addr = sd->start + sym.st_value;
				return true;
			}
		}
		return false;
	}

	void SaveAUXVData(const std::string &exe_path, int mem_fd, std::ofstream &core_file) {
		unsigned long debug_ptr = 0;
		NoteCoreAUXV *auxv_note = nullptr;

		for (auto &note : m_notes) {
			if (note->type == PT_PHDR) {
				auxv_note = static_cast<NoteCoreAUXV*>(note.get());
				break;
			}
		}
		if (!auxv_note)
			return;

		auto *auxv_data = reinterpret_cast<const typename T::auxv_t*>(auxv_note->auxv().data());

		auto phnum = AUXVval(auxv_data, AT_PHNUM);
		auto phdr = AUXVval(auxv_data, AT_PHDR);

		if (phdr == 0) {
			logger.log_info("PHDR not found for %s", exe_path);
			return;
		}

		typename T::Addr relocation = 0, dyn_addr = 0;

		size_t i;

		for (i = 0; i < phnum; i++) {
			auto addr = phdr + (sizeof(typename T::Phdr) * i) + offsetof(typename T::Phdr, p_type);
			uint32_t val;
			ReadFromFile(mem_fd, addr, &val, sizeof(val));

			if (val == PT_NULL) {
				break;
			} else if (val == PT_PHDR) {
				addr = phdr + (sizeof(typename T::Phdr) * i) + offsetof(typename T::Phdr, p_vaddr);
				ReadFromFile(mem_fd, addr, &relocation, sizeof(relocation));
				if (relocation > phdr) {
					logger.log_error("relocation is greather than phdr");
					return;
				}
				relocation = phdr - relocation;
			} else if (val == PT_DYNAMIC) {
				addr = phdr + (sizeof(typename T::Phdr) * i) + offsetof(typename T::Phdr, p_vaddr);
				ReadFromFile(mem_fd, addr, &dyn_addr, sizeof(dyn_addr));
			}
		}

		DumpData(mem_fd, core_file, phdr, sizeof(typename T::Phdr)*i, "phdr");

		SaveSymData(exe_path, relocation);

		dyn_addr += relocation;


		for (i = 0; ; i++) {
			unsigned long addr = dyn_addr + (sizeof(typename T::Dyn)*i) + offsetof(typename T::Dyn, d_tag);
			uint32_t val;
			ReadFromFile(mem_fd, addr, &val, sizeof(val));
			if (val == DT_NULL)
				break;
			if (val == DT_DEBUG) {
				addr = dyn_addr + sizeof(typename T::Dyn)*i + offsetof(typename T::Dyn, d_un.d_ptr);
				ReadFromFile(mem_fd, addr, &debug_ptr, sizeof(debug_ptr));
			}
		}
		if (!debug_ptr) {
				logger.log_error("debug_ptr not found");
				return;
		}
		DumpData(mem_fd, core_file, dyn_addr, sizeof(typename T::Dyn)*i, "auxv dyns");
		DumpData(mem_fd, core_file, debug_ptr, sizeof(struct r_debug), "auxv r_debug");

		unsigned long ptr = debug_ptr;
		ReadFromFile(mem_fd, ptr + offsetof(struct r_debug, r_map), &ptr, sizeof(ptr));

		while (ptr) {
			unsigned long addr = 0;
			DumpData(mem_fd, core_file, ptr, sizeof(struct link_map), "auxv link_map");

			ReadFromFile(mem_fd, ptr + offsetof(struct link_map, l_name), &addr, sizeof(addr));

			char buff[PAGESIZE];
			ReadFromFile(mem_fd, addr, buff, sizeof(buff));
			buff[sizeof(buff)-1] = 0;

			if (buff[0] != 0) {
				DumpData(mem_fd, core_file, addr, strnlen(buff, sizeof(buff)) + 1, "l_name");
				ReadFromFile(mem_fd, ptr+offsetof(struct link_map, l_addr), &addr, sizeof(addr));
				SaveSymData(buff, addr);
			}

			ReadFromFile(mem_fd, ptr + offsetof(struct link_map, l_next), &ptr, sizeof(ptr));
		}
	}

	unsigned long GetStackPointer(int task_nr, std::map<pid_t, user_regs_struct> &registers) {
		unsigned long result;
#if defined(__x86_64__)
		result = registers[task_nr].rsp;
#elif defined(__i386__)
		result = registers[task_nr].esp;
#elif defined(__arm__)
		result = registers[task_nr].uregs[13];
#elif defined(__aarch64__)
		result = registers[task_nr].sp;
#else
#error Unsupported architecture
#endif
		return result;
	}

	void SaveStacks(int mem_fd, std::ofstream &core_file, std::vector<pid_t> tasks, std::map<pid_t, user_regs_struct> registers) {
		for (auto task_nr : tasks) {
			unsigned long sp = GetStackPointer(task_nr, registers);
			auto *entry = FindRecordForMemAddress(sp);
			size_t len = entry->header.p_vaddr + entry->header.p_memsz - sp;
			auto out_addr = MemAddr2FileAddr(sp);
			DumpData(mem_fd, core_file, sp, out_addr, len, "stack nr " + std::to_string(task_nr));
		}
	}
};
#endif	// CORE_HPP__
